# -*- coding: utf-8 -*-

"""
Provides data access classes for accessing and managing MythTV data
"""

from MythStatic import *
from MythBase import *

import re, sys, socket, os
import xml.dom.minidom as minidom
from time import mktime, strftime
from datetime import date, datetime
from socket import gethostbyaddr, gethostname

#### FILE ACCESS ####

def ftopen(file, type, forceremote=False, nooverwrite=False, db=None):
    db = MythDBConn(db)
    log = MythLog('Python File Transfer', db=db)
    reuri = re.compile(\
        'myth://((?P<group>.*)@)?(?P<host>[a-zA-Z0-9\.]*)(:[0-9]*)?/(?P<file>.*)')
    reip = re.compile('(?:\d{1,3}\.){3}\d{1,3}')

    if type not in ('r','w'):
        raise MythError("File I/O must be of type 'r' or 'w'")

    # process URI (myth://<group>@<host>[:<port>]/<path/to/file>)
    match = reuri.match(file)
    if match is None:
        raise MythError('Invalid FileTransfer input string: '+file)
    host = match.group('host')
    filename = match.group('file')
    sgroup = match.group('group')
    if sgroup is None:
        sgroup = 'Default'

    # get full system name
    if reip.match(host):
        c = db.cursor()
        query = """SELECT hostname FROM settings
                        WHERE value='BackendServerIP' AND data=%s"""
        log.log(log.DATABASE, query, host)
        if c.execute(query, host) == 0:
            c.close()
            raise MythError('Backend not found in database: '+backend)
        host = c.fetchone()[0]
        c.close()

    # user forced to remote access
    if forceremote:
        if (type == 'w') and (filename.find('/') != -1):
            raise MythError('All remote writes must be to base path')
        return FileTransfer(host, filename, sgroup, type, db)

    sgs = db.getStorageGroup(groupname=sgroup, hostname=host)

    if type == 'w':
        # prefer local storage always
        for i in range(len(sgs)-1,-1,-1):
            if not sgs[i].local:
                sgs.pop(i)
            else:
                st = os.statvfs(sgs[i].dirname)
                sgs[i].free = st[0]*st[3]
        if len(sgs) > 1:
            # choose path with most free space
            sg = sgs.pop()
            while len(sgs):
                if sgs[0].free > sg.free: sg = sgs.pop()
                else: sgs.pop()
            # check that folder exists
            if filename.find('/') != -1:
                path = sg.dirname+filename.rsplit('/',1)[0]
                if not os.access(path, os.F_OK):
                    os.makedirs(path)
            if nooverwrite:
                if os.access(sg.dirname+filename, os.F_OK):
                    raise MythError('File already exists')
            log.log(log.FILE, 'Opening local file (w)', sg.dirname+filename)
            return open(sg.dirname+filename, 'w')
        elif len(sgs) == 1:
            if filename.find('/') != -1:
                path = sgs[0].dirname+filename.rsplit('/',1)[0]
                if not os.access(path, os.F_OK):
                    os.makedirs(path)
            if nooverwrite:
                if os.access(sgs[0].dirname+filename, os.F_OK):
                    raise MythError('File already exists')
            log.log(log.FILE, 'Opening local file (w)', sgs[0].dirname+filename)
            return open(sgs[0].dirname+filename, 'w')
        else:
            if filename.find('/') != -1:
                raise MythError('All remote writes must be to base path')
            return FileTransfer(host, filename, sgroup, 'w', db)
    else:
        # search for file in local directories
        for sg in sgs:
            if sg.local:
                if os.access(sg.dirname+filename, os.F_OK):
                    # file found, open local
                    log.log(log.FILE, 'Opening local file (r)',
                                                sg.dirname+filename)
                    return open(sg.dirname+filename, type)
        # file not found, open remote
        return FileTransfer(host, filename, sgroup, type, db)

class FileTransfer( MythBEConn ):
    """
    A connection to mythbackend intended for file transfers
    """
    logmodule = 'Python FileTransfer'
    def __repr__(self):
        return "<open file 'myth://%s:%s/%s', mode '%s' at %s>" % \
                          (self.sgroup, self.host, self.filename, \
                                self.type, hex(id(self)))

    def __init__(self, host, filename, sgroup, type, db=None):
        if type not in ('r','w'):
            raise MythError("FileTransfer type must be read ('r') "+
                                "or write ('w')")

        self.host = host
        self.filename = filename
        self.sgroup = sgroup
        self.type = type

        self.tsize = 2**15
        self.count = 0
        self.step = 2**12

        self.open = False
        # create control socket
        self.control = MythBEConn(host, 'Playback', db=db)
        self.control.log.module = 'Python FileTransfer Control'
        # continue normal Backend initialization
        MythBEConn.__init__(self, host, type, db=db, single=True)
        self.open = True

    def announce(self, type):
        # replacement announce for Backend object
        if type == 'w':
            self.w = True
        elif type == 'r':
            self.w = False
        res = self.backendCommand('ANN FileTransfer %s %d %d %s' \
                    % (socket.gethostname(), self.w, False,
                        BACKEND_SEP.join(['-1',self.filename,self.sgroup])))
        if res.split(BACKEND_SEP)[0] != 'OK':
            raise MythError('Unexpected answer to ANN command: %s' % res)
        else:
            sp = res.split(BACKEND_SEP)
            self.sockno = int(sp[1])
            self.pos = 0
            self.size = (int(sp[2]) + (int(sp[3])<0))*2**32 + int(sp[3])

    def __del__(self):
        if not self.open:
            return
        self.control.backendCommand('QUERY_FILETRANSFER '\
                    +BACKEND_SEP.join([str(self.sockno), 'JOIN']))
        self.socket.shutdown(1)
        self.socket.close()
        self.open = False

    def tell(self):
        """
        Return the current offset from the beginning of the file
        """
        return self.pos

    def close(self):
        """
        Close the data transfer socket
        """
        self.__del__()

    def rewind(self):
        """
        Seek back to the start of the file
        """
        self.seek(0)

    def read(self, size):
        """
        Read a block of data, requests over 64KB will be buffered internally
        """

        def recv(self, size):
            buff = ''
            while len(buff) < size:
                buff += self.socket.recv(size-len(buff))
            return buff

        if self.w:
            print 'Error: attempting to read from a write-only socket'
            return
        if size == 0:
            return ''

        final = self.pos+size
        if final > self.size:
            final = self.size

        buff = ''
        while self.pos < final:
            self.count += 1
            size = final - self.pos
            if size > self.tsize:
                size = self.tsize
            res = self.control.backendCommand('QUERY_FILETRANSFER '+\
                        BACKEND_SEP.join([  str(self.sockno),
                                            'REQUEST_BLOCK',
                                            str(size)]))
            #print '%s - %s' % (int(res), size)
            if int(res) == size:
                if (self.count == 10) and (self.tsize < 2**17) :
                    self.count = 0
                    self.tsize += self.step
            else:
                if int(res) == -1:
                    self.seek(self.pos)
                    continue
                size = int(res)
                self.count = 0
                self.tsize -= self.step
                if self.tsize < self.step:
                    self.tsize = self.step

            buff += recv(self, size)
            self.pos += size
        return buff

    def write(self, data):
        """
        Write a block of data, requests over 64KB will be buffered internally
        """
        if not self.w:
            print 'Error: attempting to write to a read-only socket'
            return
        while len(data) > 0:
            size = len(data)
            if size > self.tsize:
                buff = data[self.tsize:]
                data = data[:self.tsize]
            else:
                buff = data
                data = ''
            self.pos += int(self.socket.send(data))
            self.control.backendCommand('QUERY_FILETRANSFER '+BACKEND_SEP\
                    .join([str(self.sockno),'WRITE_BLOCK',str(size)]))
        return


    def seek(self, offset, whence=0):
        """
        Seek 'offset' number of bytes
           whence==0 - from start of file
                   whence==1 - from current position
        """
        if whence == 0:
            if offset < 0:
                offset = 0
            if offset > self.size:
                offset = self.size
        elif whence == 1:
            if offset + self.pos < 0:
                offset = -self.pos
            if offset + self.pos > self.size:
                offset = self.size - self.pos
        elif whence == 2:
            if offset > 0:
                offset = 0
            if offset < -self.size:
                offset = -self.size

        curhigh,curlow = self.splitInt(self.pos)
        offhigh,offlow = self.splitInt(offset)

        res = self.control.backendCommand('QUERY_FILETRANSFER '+BACKEND_SEP\
                        .join([str(self.sockno),'SEEK',str(offhigh),
                                str(offlow),str(whence),str(curhigh),
                                str(curlow)])\
                    ).split(BACKEND_SEP)
        self.pos = self.joinInt(int(res[0]),int(res[1]))

class FileOps( MythBEConn ):
    """
    Several canned file management tasks.
    """
    logmodule = 'Python Backend FileOps'

    def getRecording(self, chanid, starttime):
        """
        Returns a Program object matching the channel id and start time
        """
        res = self.backendCommand('QUERY_RECORDING TIMESLOT %d %d' \
                        % (chanid, starttime)).split(BACKEND_SEP)
        if res[0] == 'ERROR':
            return None
        else:
            return Program(res[1:], db=self.db)

    def deleteRecording(self,program,force=False):
        """
        Deletes recording, set 'force' true if file cannot be found
        Returns the number fewer recordings that exist afterwards
        """
        command = 'DELETE_RECORDING'
        if force:
            command = 'FORCE_DELETE_RECORDING'
        return self.backendCommand(BACKEND_SEP.join(\
                    [command,program.toString()]))

    def forgetRecording(self,program):
        """
        Forgets old recording and allows it to be re-recorded
        """
        self.backendCommand(BACKEND_SEP.join(['FORGET_RECORDING',
                    program.toString()]))

    def deleteFile(self,file,sgroup):
        """
        Deletes a file from specified storage group on the connected backend
        Takes a relative file path from the root of the storage group
        Returns 1 on success
        """
        return self.backendCommand(BACKEND_SEP.join(\
                    ['DELETE_FILE',file,sgroup]))

    def getHash(self, file, sgroup):
        """
        Returns a hash calculated from the first and last 128K, and filesize
        Must be run on the backend housing the files.
        """
        return self.backendCommand(BACKEND_SEP.join((\
                    'QUERY_FILE_HASH',file, sgroup)))

class FreeSpace( DictData ):
    """
    Represents a Storage Group entry
    """
    field_order = [ 'host',         'path',     'islocal',
                    'disknumber',   'sgroupid', 'blocksize',
                    'ts_high',      'ts_low',   'us_high',
                    'us_low']
    field_type = [3, 3, 2, 0, 0, 0, 0, 0, 0, 0]
    def __str__(self):
        return "Storage Group %d (%s@%s)"\
                    % (self.sgroupid, self.path, self.host)

    def __repr__(self):
        return "Storage Group %d (%s@%s)"\
                    % (self.sgroupid, self.path, self.host)

    def __init__(self, raw):
        DictData.__init__(self, raw)
        self.totalspace = self.joinInt(self.ts_high, self.ts_low)
        self.usedspace = self.joinInt(self.us_high, self.us_low)
        self.freespace = self.totalspace - self.usedspace


#### RECORDING ACCESS ####

class Program( DictData ):
    """
    Represents a program with all the detail known.
    """

    field_order = [ 'title',        'subtitle',     'description',
                    'category',     'chanid',       'channum',
                    'callsign',     'channame',     'filename',
                    'fs_high',      'fs_low',       'starttime',
                    'endtime',      'duplicate',    'shareable',
                    'findid',       'hostname',     'sourceid',
                    'cardid',       'inputid',      'recpriority',
                    'recstatus',    'recordid',     'rectype',
                    'dupin',        'dupmethod',    'recstartts',
                    'recendts',     'repeat',       'programflags',
                    'recgroup',     'commfree',     'outputfilters',
                    'seriesid',     'programid',    'lastmodified',
                    'stars',        'airdate',      'hasairdate',
                    'playgroup',    'recpriority2', 'parentid',
                    'storagegroup', 'audio_props',  'video_props',
                    'subtitle_type','year']
    field_type = [  3,      3,      3,
                    3,      0,      3,
                    3,      3,      3,
                    0,      0,      4,
                    4,      0,      0,
                    0,      3,      0,
                    0,      0,      0,
                    0,      0,      3,
                    0,      0,      4,
                    4,      0,      3,
                    3,      0,      3,
                    3,      3,      3,
                    1,      3,      0,
                    3,      0,      3,
                    3,      0,      0,
                    0,      0]
    def __str__(self):
        return u"<Program '%s','%s' at %s" % (self.title,
                 self.starttime.strftime('%Y-%m-%d %H:%M:%S'), hex(id(self)))

    def __repr__(self):
        return str(self).encode('utf-8')

    def __init__(self, raw, db=None):
        self.db = MythDBConn(db)
        DictData.__init__(self, raw)
        self.filesize = self.joinInt(self.fs_high,self.fs_low)

    def toString(self):
        data = []
        for i in range(0,PROGRAM_FIELDS):
            if self.data[self.field_order[i]] == None:
                datum = ''
            elif self.field_type[i] in (0,1,3):
                datum = str(self.data[self.field_order[i]])
            elif self.field_type[i] == 2:
                datum = str(int(self.data[self.field_order[i]]))
            elif self.field_type[i] == 4:
                datum = self.data[self.field_order[i]].strftime('%Y%m%d%H%M%S')
            data.append(datum)
        return BACKEND_SEP.join(data)

    def delete(self, rerecord=False):
        be = FileOps(db=self.db)
        be.deleteRecording(self)
        if rerecord:
            be.forgetRecording(self)

    def getRecorded(self):
        """
        Return a Recorded object for manipulation of recording
        """
        return Recorded((self.chanid,self.recstartts), db=self.db)

    def open(self, type='r'):
        if type != 'r':
            MythError('Program() only supports read-only access to files')
        return ftopen(self.filename, 'r')


class Recorded( DBDataWrite ):
    """
    Class for managing recordings
    """
    table = 'recorded'
    where = 'chanid=%s AND starttime=%s'
    setwheredat = 'self.chanid,self.starttime'
    defaults = {'title':u'Unknown', 'subtitle':'',          'description':'',
                'category':'',      'hostname':'',          'bookmark':0,
                'editing':0,        'cutlist':0,            'autoexpire':1,
                'commflagged':0,    'recgroup':'Default',   'recordid':0,
                'seriesid':'',      'programid':'',         'lastmodified':'',
                'filesize':0,       'stars':0,              'previouslyshown':0,
                'originalairdate':'',                       'preserve':0,
                'findid':0,         'deletepending':0,      'transcoder':56,
                'timestretch':1,    'recpriority':0,        'playgroup':'Default',
                'profile':'Default','duplicate':1,          'transcoded':0,
                'watched':0,        'storagegroup':'Default'}
    logmodule = 'Python Recorded'

    class _Cast( DBDataCRef ):
        table = 'people'
        rtable = 'recordedcredits'
        t_ref = 'person'
        t_add = ['name']
        r_ref = 'person'
        r_add = ['role']
        w_field = ['chanid','starttime']
        def __repr__(self):
            if self.data is None:
                self._fill()
            if len(self.data) == 0:
                return 'No cast'
            cast = []
            for member in self.data:
                cast.append(member.name)
            return ', '.join(cast).encode('utf-8')

    class _Seek( DBDataRef ):
        table = 'recordedseek'
        wfield = ['chanid','starttime']

    class _Markup( DBDataRef ):
        table = 'recordedmarkup'
        wfield = ['chanid','starttime']
        
    def __str__(self):
        return u"<Recorded '%s','%s' at %s>" % (self.title,
                self.starttime.strftime('%Y-%m-%d %H:%M:%S'), hex(id(self)))

    def __repr__(self):
        return str(self).encode('utf-8')

    def __init__(self, data=None, db=None, raw=None):
        """
        Creates a new Recorded object. Supply database object if possible.

        Can be initialized with either variable:        
            data = (chanid, starttime)
            raw = row returned from 'select * from recorded'
        Can be left empty to allow creation of new database entry
        """
        DBDataWrite.__init__(self, data, db, raw)
        if (data is not None) or (raw is not None):
            self.cast = self._Cast(self.wheredat, self.db)
            self.seek = self._Seek(self.wheredat, self.db)
            self.markup = self._Markup(self.wheredat, self.db)

    def create(self, data=None):
        """
        Creates new database entry using stored data. Can be overridden by 
        optional 'data' input as dictionary
        """
        DBDataWrite.create(self, data)
        self.wheredat = (self.chanid,self.starttime)
        self._pull()
        self.cast = self._Cast(self.wheredat, self.db)
        return self

    def delete(self, rerecord=False):
        """Triggers deletion of recording on backend"""
        self.getProgram().delete(rerecord)

    def open(self, type='r'):
        """Open file pointer to referenced recording"""
        return ftopen("myth://%s@%s/%s" % ( self.storagegroup, \
                                            self.hostname,\
                                            self.basename), type, db=self.db)

    def getProgram(self):
        """Return matching Program object"""
        be = FileOps(db=self.db)
        return be.getRecording(self.chanid, 
                    int(self.starttime.strftime('%Y%m%d%H%M%S')))

    def getRecordedProgram(self):
        """Return matching RecordedProgram object"""
        return RecordedProgram((self.chanid,self.progstart), db=self.db)

    def formatPath(self, path, replace=None):
        for (tag, data, format) in (('T','title','%s'),('S','subtitle','%s'),
                             ('R','description','%s'), ('C','category','%s'),
                             ('U','recgroup','%s'),    ('hn','hostname','%s'),
                             ('c','chanid','%d') ):
            path = path.replace('%'+tag, format % self[data])
        for (data, pre) in (   ('starttime','%'), ('endtime','%e'),
                               ('progstart','%p'),('progend','%pe') ):
            for (tag, format) in (('y','%y'),('Y','%Y'),('n','%m'),('m','%m'),
                                  ('j','%d'),('d','%d'),('g','%I'),('G','%H'),
                                  ('h','%I'),('H','%H'),('i','%M'),('s','%S'),
                                  ('a','%p'),('A','%p') ):
                path = path.replace(pre+tag, self[data].strftime(format))
        for (tag, format) in (('y','%y'),('Y','%Y'),('n','%m'),('m','%m'),
                              ('j','%d'),('d','%d')):
            path = path.replace('%o'+tag,
                    self['originalairdate'].strftime(format))
        path = path.replace('%-','-')
        path = path.replace('%%','%')
        path += '.'+self['basename'].split('.')[-1]

        # clean up for windows
        if replace is not None:
            for char in ('\\',':','*','?','"','<','>','|'):
                path = path.replace(char, replace)
        return path

class RecordedProgram( DBDataWrite ):
    """
    Class for managing additional data for Recorded
    """
    table = 'recordedprogram'
    where = 'chanid=%s AND starttime=%s'
    setwheredat = 'self.chanid,self.starttime'
    defaults = {'title':'',     'subtitle':'',          'description':'',
                'category':'',  'category_type':'',     'airdate':'',
                'stars':0,      'previouslyshown':'',   'title_pronounce':'',
                'stereo':0,     'subtitled':0,          'hdtv':0,
                'partnumber':0, 'closecaptioned':0,     'parttotal':0,
                'seriesid':'',  'originalairdate':'',   'showtype':u'Series',
                'colorcode':'', 'syndicateepisodenumber':'',
                'programid':'', 'manualid':0,           'generic':0,
                'first':0,      'listingsource':0,      'last':0,
                'audioprop':'', 'subtitletypes':'',     'videoprop':''}
    logmodule = 'Python RecordedProgram'

    def __str__(self):
        return u"<RecordedProgram '%s','%s' at %s>" % (self.title,
                self.starttime.strftime('%Y-%m-%d %H:%M:%S'), gex(id(self)))

    def __repr__(self):
        return str(self).encode('utf-8')

    def __init__(self, data=None, db=None, raw=None):
        """
        Creates a new RecordedProgram object. Supply database object if possible.

        Can be initialized with either variable:        
            data = (chanid, starttime)
            raw = row returned from 'select * from recorded'
        Can be left empty to allow creation of new database entry
        """
        DBDataWrite.__init__(self, data, db, raw)

    def create(self, data=None):
        """
        Creates new database entry using stored data. Can be overridden by 
        optional 'data' input as dictionary
        """
        DBDataWrite.create(self, data)
        self.wheredat = (self.chanid, self.starttime)
        self._pull()
        return self

class OldRecorded( DBDataWrite ):
    """
    Class for managing old recordings.
    """
    table = 'oldrecorded'
    where = 'chanid=%s AND starttime=%s'
    setwheredat = 'self.chanid,self.starttime'
    defaults = {'title':'',     'subtitle':'',      'description':'',
                'category':'',  'seriesid':'',      'programid':'',
                'findid':0,     'recordid':0,       'station':'',
                'rectype':1,    'duplicate':1,      'recstatus':-3,
                'reactivate':0, 'generic':0}
    logmodule = 'Python OldRecorded'

    def __str__(self):
        return u"<OldRecorded '%s','%s' at %s>" % (self.title,
                self.starttime.strftime('%Y-%m-%d %H:%M:%S'), gex(id(self)))

    def __repr__(self):
        return str(self).encode('utf-8')

    def __init__(self, data=None, db=None, raw=None):
        """
        Creates a new OldRecorded object. Supply database object if possible.

        Can be initialized with either variable:        
            data = (chanid, starttime)
            raw = row returned from 'select * from recorded'
        Can be left empty to allow creation of new database entry
        """
        DBDataWrite.__init__(self, data, db, raw)
        
    def create(self, data=None):
        """
        Creates new database entry using stored data. Can be overridden by 
        optional 'data' input as dictionary
        """
        DBDataWrite.create(self, data)
        self.wheredat = (self.chanid, self.starttime)
        self._pull()
        return self

    def setDuplicate(self, record=False):
        c = self.db.cursor()
        c.execute("""UPDATE oldrecorded SET duplicate=%%s
                     WHERE %s""" % self.where, \
                tuple([record]+list(self.wheredat)))

    def update(self, *args, **keywords):
        """OldRecorded entries can not be altered"""
        return
    def delete(self):
        """OldRecorded entries cannot be deleted"""
        return

class Job( DBDataWrite ):
    """
    Class for managing tasks within mythtv's jobqueue
    """
    # types
    NONE         = 0x0000
    SYSTEMJOB    = 0x00ff
    TRANSCODE    = 0x0001
    COMMFLAG     = 0x0002
    USERJOB      = 0xff00
    USERJOB1     = 0x0100
    USERJOB2     = 0x0200
    USERJOB3     = 0x0300
    USERJOB4     = 0x0400
    # cmds
    RUN          = 0x0000
    PAUSE        = 0x0001
    RESUME       = 0x0002
    STOP         = 0x0004
    RESTART      = 0x0008
    # flags
    NO_FLAGS     = 0x0000
    USE_CUTLIST  = 0x0001
    LIVE_REC     = 0x0002
    EXTERNAL     = 0x0004
    # statuses
    UNKNOWN      = 0x0000
    QUEUED       = 0x0001
    PENDING      = 0x0002
    STARTING     = 0x0003
    RUNNING      = 0x0004
    STOPPING     = 0x0005
    PAUSED       = 0x0006
    RETRY        = 0x0007
    ERRORING     = 0x0008
    ABORTING     = 0x0008
    DONE         = 0x0100
    FINISHED     = 0x0110
    ABORTED      = 0x0120
    ERRORED      = 0x0130
    CANCELLED    = 0x0140

    table = 'jobqueue'
    logmodule = 'Python Jobqueue'
    defaults = {'id': None}

    def __str__(self):
        return u"<Job '%s' at %s>" % (self.id, hex(id(self)))

    def __repr__(self):
        return str(self).encode('utf-8')

    def __init__(self, id=None, chanid=None, starttime=None, \
                        db=None, raw=None):
        """
        Creates a new job object. Supply with database object when possible.

        Can be initialized by setting:
            jobid --or-- chanid AND starttime --or-- raw
        Can be left empty to allow creation of new job.
        """
        self.__dict__['where'] = 'id=%s'
        self.__dict__['setwheredat'] = 'self.id,'

        if raw is not None:
            DBDataWrite.__init__(self, None, db, raw)
        elif id is not None:
            DBDataWrite.__init__(self, (id,), db, None)
        elif (chanid is not None) and (starttime is not None):
            self.__dict__['where'] = 'chanid=%s AND starttime=%s'
            DBDataWrite.__init__(self, (chanid,starttime), db, None)

    def create(self, data):
        """
        Creates new database entry using stored data. Can be overridden by 
        optional 'data' input as dictionary
        """
        id = DBDataWrite.create(self, data)
        self.where = 'id=%s'
        self.wheredat = (id,)
        return self

    def setComment(self,comment):
        self.comment = comment
        self.update()

    def setStatus(self,status):
        self.status = status
        self.update()

class Channel( DBDataWrite ):
    """
    Represents a single channel from the channel table
    """
    table = 'channel'
    where = 'chanid=%s'
    setwheredat = 'self.chanid,'
    defaults = {'icon':'',              'finetune':0,       'videofilters':'',
                'xmltvid':'',           'recpriority':0,    'contrast':32768,
                'brightness':32768,     'colour':32768,     'hue':32768,
                'tvformat':u'Default',  'visible':1,        'outputfilters':'',
                'useonairguide':0,      'mplexid':32767,    'serviceid':0,
                'tmoffset':0,           'default_authority':'',
                'commmethod':-1}
    logmodule = 'Python Channel'
    
    def __str__(self):
        return u"<Channel '%s','%s' at %s>" % \
                        (self.chanid, self.name, hex(id(self)))

    def __repr__(self):
        return str(self).encode('utf-8')

    def __init__(self, chanid=None, db=None, raw=None):
        """
        Creates a new Channel object. Supply database object if possible.

        Can be initialized by setting:
            chanid --or-- raw
        Can be left empty to allow creation of new database entry
        """
        DBDataWrite.__init__(self, chanid, db, raw)

    def create(self, data):
        """
        Creates new database entry using stored data. Can be overridden by 
        optional 'data' input as dictionary
        """
        DBDataWrite.create(self, data)
        self.wheredat = (data.chanid)
        self._pull()
        return self
    
#    def openIcon(self, mode='r'):
    

class Guide( DBData ):
    """
    Represents a single program from the program guide
    """
    table = 'program'
    where = 'chanid=%s AND starttime=%s'
    setwheredat = 'self.chanid,self.starttime'
    logmodule = 'Python Guide'
    
    def __str__(self):
        return u"<Guide '%s','%s' at %s>" % (self.title,
                self.starttime.strftime('%Y-%m-%d %H:%M:%S'), hex(id(self)))

    def __repr__(self):
        return str(self).encode('utf-8')


#### MYTHVIDEO ####

class Video( DBDataWrite ):
    """
    Class for managing and creating metadata for a single video file
    """
    table = 'videometadata'
    where = 'intid=%s'
    setwheredat = 'self.intid,'
    defaults = {'subtitle':u'',             'director':u'Unknown',
                'rating':u'NR',             'inetref':u'00000000',
                'year':1895,                'userrating':0.0,
                'length':0,                 'showlevel':1,
                'coverfile':u'No Cover',    'host':u'',
                'intid':None,               'homepage':u'',
                'watched':False,            'category':0,
                'browse':True,              'hash':0,
                'season':0,                 'episode':0,
                'releasedate':date(1,1,1),  'childid':0,
                'insertdate': datetime.now()}
    logmodule = 'Python Video'
    schema_value = 'mythvideo.DBSchemaVer'
    schema_local = MVSCHEMA_VERSION
    schema_name = 'MythVideo'

    def __repr__(self):
        return str(self).encode('utf-8')

    def __str__(self):
        res = self.title
        if self.season and self.episode:
            res += u' - %dx%02d' % (self.season, self.episode)
        if self.subtitle:
            res += u' - '+self.subtitle
        return u"<Video '%s' at %s>" % (res, hex(id(self)))

    def __init__(self, id=None, db=None, raw=None):
        """
        Creates a new Video object. Supply database object if possible.

        Can be initialized by setting:        
            id --or-- raw
        Can be left empty to allow creation of new database entry
        """
        DBDataWrite.__init__(self, (id,), db, raw)
        if (id is not None) or (raw is not None):
            self.cast = self._Cast((self.intid,), self.db)
            self.genre = self._Genre((self.intid,), self.db)
            self.country = self._Country((self.intid,), self.db)

    def create(self, data=None):
        """
        Creates new database entry using stored data. Can be overridden by 
        optional 'data' input as dictionary
        """
        c = self.db.cursor()
        fields = ' AND '.join(['%s=%%s' % f for f in \
                        ('title','subtitle','season','episode')])
        count = c.execute("""SELECT inetref FROM videometadata WHERE %s""" %
                fields, (self.title, self.subtitle, self.season, self.episode))
        if count:
            id = c.fetchone()[0]
        else:
            id = DBDataWrite.create(self, data)
        c.close()
        self.wheredat = (id,)
        self._pull()
        self.cast = self._Cast((self.intid,), self.db)
        self.genre = self._Genre((self.intid,), self.db)
        self.country = self._Country((self.intid,), self.db)
        return self

    class _Cast( DBDataCRef ):
        table = 'videocast'
        rtable = 'videometadatacast'
        t_ref = 'intid'
        t_add = ['cast']
        r_ref = 'idcast'
        r_add = []
        w_field = ['idvideo']
        def __repr__(self):
            if self.data is None:
                self._fill()
            if len(self.data) == 0:
                return ''
            cast = []
            for member in self.data:
                cast.append(member.cast)
            return u', '.join(cast).encode('utf-8')
        def clean(self):
            for member in self.data:
                self.delete(member.cast)

    class _Genre( DBDataCRef ):
        table = 'videogenre'
        rtable = 'videometadatagenre'
        t_ref = 'intid'
        t_add = ['genre']
        r_ref = 'idgenre'
        r_add = []
        w_field = ['idvideo']
        def __repr__(self):
            if self.data is None:
                self._fill()
            if len(self.data) == 0:
                return ''
            genre = []
            for member in self.data:
                genre.append(member.genre)
            return ', '.join(genre).encode('utf-8')
        def clean(self):
            for member in self.data:
                self.delete(member.genre)

    class _Country( DBDataCRef ):
        table = 'videocountry'
        rtable = 'videometadatacountry'
        t_ref = 'intid'
        t_add = ['country']
        r_ref = 'idcountry'
        r_add = []
        w_field = ['idvideo']
        def __repr__(self):
            if self.data is None:
                self._fill()
            if len(self.data) == 0:
                return ''
            country = []
            for member in self.data:
                country.append(member.country)
            return ', '.join(country).encode('utf-8')
        def clean(self):
            for member in self.data:
                self.delete(member.country)

    class _Markup( DBDataRef ):
        table = 'filemarkup'
        wfield = ['filename',]

    def _open(self, type, mode='r',nooverwrite=False):
        """
        Open file pointer
        """
        sgroup = {  'filename':'Videos',        'banner':'Banners',
                    'coverfile':'Coverart',     'fanart':'Fanart',
                    'screenshot':'Screenshots', 'trailer':'Trailers'}
        if self.data is None:
            return None
        if type not in sgroup:
            raise MythError('Video._open() - '+\
                            'Trying to open invalid file type: '+str(type))
        SG = self.db.getStorageGroup(sgroup[type], self.host)
        if len(SG) == 0:
            SG = self.db.getStorageGroup('Videos', self.host)
            if len(SG) == 0:
                raise MythError('Could not find storagegroup')
        return ftopen('myth://%s@%s/%s' % ( SG[0].groupname,
                                            self.host,
                                            self.data[type]),
                            mode, False, nooverwrite, self.db)

    def delete(self):
        """
        Delete video entry from database (does not delete file)
        """
        if self.data is None:
            return
        self.cast.clean()
        self.genre.clean()
        self.country.clean()
        DBDataWrite.delete(self)

    def open(self,mode='r',nooverwrite=False):
        """Open file pointer to video"""
        return self._open('filename',mode,nooverwrite)

    def openBanner(self,mode='r',nooverwrite=False):
        """Open file pointer to banner image"""
        return self._open('banner',mode,nooverwrite)

    def openCoverart(self,mode='r',nooverwrite=False):
        """Open file pointer to coverart image"""
        return self._open('coverfile',mode,nooverwrite)

    def openFanart(self,mode='r',nooverwrite=False):
        """Open file pointer to fanart image"""
        return self._open('fanart',mode,nooverwrite)

    def openScreenshot(self,mode='r',nooverwrite=False):
        """Open file pointer to screenshot image"""
        return self._open('screenshot',mode,nooverwrite)

    def openTrailer(self,mode='r',nooverwrite=False):
        """Open file pointer to trailer"""
        return self._open('trailer',mode,nooverwrite)

    def getHash(self):
        """Returns computed file hash"""
        if self.host is None:
            return None
        be = FileOps(self.host)
        hash = be.getHash(self.filename, 'Videos')
        be.close()
        return hash

class VideoGrabber( Grabber ):
    """Customized Grabber class specifically for MythVideo"""
    logmodule = 'Python MythVideo Grabber'

    def __init__(self, mode, lang='en', db=None):
        """Accepts 'mode' as 'TV' or 'Movie', and an optional language"""
        if mode == 'TV':
            Grabber.__init__(self, setting='mythvideo.TVGrabber', db=db)
        elif mode == 'Movie':
            Grabber.__init__(self, setting='mythvideo.MovieGrabber', db=db)
        else:
            raise MythError('Invalid MythVideo grabber')
        self._check_schema('mythvideo.DBSchemaVer',
                                MVSCHEMA_VERSION, 'MythVideo')
        self.append('-l',lang)

    def searchTitle(self, title, year=None):
        regex = re.compile('([0-9]+):(.+?)( \(([0-9]{4})\)\n|\n)')
        res = self.command('-M', '"%s"' %title)
        ret = []
        for m in regex.finditer(res):
            m = list(m.group(1,2,4))
            m[0] = int(m[0])
            if m[2]:
                m[2] = int(m[2])
            if year and m[2]:
                if m[2] != int(year):
                    continue
            ret.append(tuple(m))
        return tuple(ret)

    def searchEpisode(self, title, subtitle):
        regex = re.compile("S(?P<season>[0-9]*)E(?P<episode>[0-9]*)")
        res = self.command('-N', '"%s"' % title, '"%s"' % subtitle)
        match = regex.match(res)
        if match:
            season = int(match.group('season'))
            episode = int(match.group('episode'))
            return (season, episode)
        else:
            return (None, None)

    def getData(self, inetref, season=None, episode=None):
        if season and episode:
            res = self.command('-D', inetref, season, episode)
        else:
            res = self.command('-D', str(inetref) )
        trans = {   'Title':'title',        'Subtitle':'subtitle',
                    'Year':'year',          'ReleaseDate':'releasedate',
                    'InetRef':'inetref',    'URL':'homepage',
                    'Director':'director',  'Plot':'plot',
                    'UserRating':'stars',   'MovieRating':'rating',
                    'Runtime':'length',     'Coverart':'coverfile',
                    'Fanart':'fanart',      'Banner':'banner',
                    'Screenshot':'screenshot'}
        dat = {}
        cast = ()
        genre = ()
        country = ()
        for point in res.split('\n')[:-1]:
            if point.find(':') == -1:
                continue
            key,val = point.split(':',1)
            if key in trans:
                dat[trans[key]] = val
            elif key == 'Cast':
                cast = tuple(val.split(', '))
            elif key == 'Genre':
                genre = tuple(val.split(', '))
            elif key == 'Country':
                country = typle(val.split(', '))
        if 'releasedate' in dat:
            dat['releasedate'] = datetime.strptime(dat['releasedate'],\
                                                '%Y-%m-%d').date()
        return (dat, cast, genre, country)

#### MYTHNETVISION ####

class NetVisionRSSItems( DBData ):
    """
    Represents a single program from the netvisionrssitems table
    """
    table = 'netvisionrssitems'
    where = 'feedtitle=%s AND title=%s'
    setwheredat = 'self.feedtitle,self.title'
    schema_value = 'NetvisionDBSchemaVer'
    schema_local = NVSCHEMA_VERSION
    schema_name = 'NetVision'
    
    def __str__(self):
        return u"<NetVisionRSSItems '%s@%s' at %s>" % \
                    (self.title, self.feedtitle, hex(id(self)))

    def __repr__(self):
        return str(self).encode('utf-8')

class NetVisionTreeItems( DBData ):
    """
    Represents a single program from the netvisiontreeitems table
    """
    table = 'netvisiontreeitems'
    where = 'feedtitle=%s AND path=%s'
    setwheredat = 'self.feedtitle,self.path'
    schema_value = 'NetvisionDBSchemaVer'
    schema_local = NVSCHEMA_VERSION
    schema_name = 'NetVision'

    def __str__(self):
        return u"<NetVisionTreeItems '%s@%s' at %s>" % \
                    (self.path, self.feedtitle, hex(id(self)))

    def __repr__(self):
        return str(self).encode('utf-8')

class NetVisionSites( DBData ):
    """
    Represents a single site from the netvisionsites table
    """
    table = 'netvisionsites'
    where = 'name=%'
    setwheredat = 'name,'
    schema_value = 'NetvisionDBSchemaVer'
    schema_local = NVSCHEMA_VERSION
    schema_name = 'NetVision'

    def __str__(self):
        return u"<NetVisionSites '%s','%s' at %s>" % \
                    (self.name, self.url, hex(id(self)))

    def __repr__(self):
        return str(self).encode('utf-8')

class NetVisionGrabber( Grabber ):
    logmodule = 'Python MythNetVision Grabber'

    @staticmethod
    def grabberList(type='search,tree', db=None):
        db = MythDBConn(db)
        self._check_schema('NetvisionDBSchemaVer',
                                NVSCHEMA_VERSION, 'NetVision')
        c = db.cursor()
        log = MythLog('Python MythNetVision Grabber', db=db)
        host = gethostname()
        glist = []
        for t in type.split(','):
            query = """SELECT name,commandline
                        FROM netvision%sgrabber
                        WHERE host=%%s""" % type
            log.log(log.DATABASE, query, host)
            c.execute(query, (host,))
            for name,commandline in c.fetchall():
                glist.append(NetVisionGrabber(name, t, db, commandline))
        c.close()
        return tuple(glist)

    def __init__(self, name, type, db=None, commandline=None):
        if type not in ('search','tree'):
            raise MythError('Invalid NetVisionGrabber() type')
        self.type = type
        self.name = name
        if commandline:
            Grabber.__init__(path=commandline, db=db)
        else:
            db = MythDBConn(db)
            self.log = MythLog(self.logmodule, db=self)
            query = """SELECT commandline
                       FROM netvision%sgrabbers
                       WHERE name=%%s AND host=%%s""" % type
            self.log.log(MythLog.DATABASE, query, str((name, gethostname())))
            if c.execute(query, (name, gethostname())) == 1:
                Grabber.__init__(path=c.fetchone()[0], db=db)
                c.close()
            else:
                c.close()
                raise MythError('NetVisionGrabber not found in database')

    def searchXML(self, title, page=1):
        return minidom.parseString(self.command('-p',page,'-S','"%s"' %title))

    def treeXML(self):
        return minidom.parseString(self.command('-T'))

    def setUpdated(self):
        if self.type is not 'tree':
            raise MythError('Can only update tree-type grabbers')
        c = db.cursor()
        c.execute("""UPDATE netvision%sgrabbers SET update=NOW()
                     WHERE name=%%s AND host=%%s""" % type,
                     (self.name, gethostname()))
        c.close()


